@@ -25,6 +25,8 @@
 
 #include "fm-config.h"
 #include "fm-archiver.h"
+#include "fm-app-info.h"
+#include "fm-utils.h"
 #include <gio/gdesktopappinfo.h>
 #include <string.h>
 
@@ -61,6 +63,7 @@ static gboolean launch_program(FmArchiver* archiver, GAppLaunchContext* ctx, con
     char* _cmd = NULL;
     const char* dir_place_holder;
     GKeyFile* dummy;
+    char* tmp;
 
     if(dir && (dir_place_holder = strstr(cmd, "%d")))
     {
@@ -70,15 +73,22 @@ static gboolean launch_program(FmArchiver* archiver, GAppLaunchContext* ctx, con
             dir_str = fm_path_to_uri(dir);
         else
         {
-            char* quote;
             GFile* gf = fm_path_to_gfile(dir);
             /* FIXME: convert dir to fuse-based local path if needed. */
             dir_str = g_file_get_path(gf);
             g_object_unref(gf);
-            quote = g_shell_quote(dir_str);
-            g_free(dir_str);
-            dir_str = quote;
         }
+
+        /* replace all % with %% so encoded URI can be handled correctly when parsing Exec key. */
+        tmp = fm_strdup_replace(dir_str, "%", "%%");
+        g_free(dir_str);
+        dir_str = tmp;
+
+        /* quote the path or URI */
+        tmp = g_shell_quote(dir_str);
+        g_free(dir_str);
+        dir_str = tmp;
+
         len = strlen(cmd) - 2 + strlen(dir_str) + 1;
         _cmd = g_malloc(len);
         len = (dir_place_holder - cmd);
@@ -93,23 +103,25 @@ static gboolean launch_program(FmArchiver* archiver, GAppLaunchContext* ctx, con
     dummy = g_key_file_new();
     g_key_file_set_string(dummy, G_KEY_FILE_DESKTOP_GROUP, "Type", "Application");
     g_key_file_set_string(dummy, G_KEY_FILE_DESKTOP_GROUP, "Name", archiver->program);
-    g_key_file_set_string(dummy, G_KEY_FILE_DESKTOP_GROUP, "Exec", cmd);
 
-    app = g_desktop_app_info_new_from_keyfile(dummy);
+    /* replace all % with %% so encoded URI can be handled correctly when parsing Exec key. */
+    g_key_file_set_string(dummy, G_KEY_FILE_DESKTOP_GROUP, "Exec", cmd);
+    app = (GAppInfo*)g_desktop_app_info_new_from_keyfile(dummy);
 
     g_key_file_free(dummy);
     g_debug("cmd = %s", cmd);
     if(app)
     {
         GList* uris = NULL, *l;
-        for(l = fm_list_peek_head_link(files); l; l=l->next)
+        for(l = fm_path_list_peek_head_link(files); l; l=l->next)
         {
             FmPath* path = FM_PATH(l->data);
             uris = g_list_prepend(uris, fm_path_to_uri(path));
         }
-        g_app_info_launch_uris(app, uris, ctx, NULL);
+        fm_app_info_launch_uris(app, uris, ctx, NULL);
         g_list_foreach(uris, (GFunc)g_free, NULL);
         g_list_free(uris);
+        g_object_unref(app);
     }
     g_free(_cmd);
     return TRUE;
@@ -182,7 +194,7 @@ void fm_archiver_set_default(FmArchiver* archiver)
 }
 
 /* get a list of FmArchiver* of all GUI archivers known to libfm */
-GList* fm_archiver_get_all()
+const GList* fm_archiver_get_all()
 {
     return archivers;
 }
@@ -192,11 +204,11 @@ void _fm_archiver_init()
     GKeyFile *kf = g_key_file_new();
     if(g_key_file_load_from_file(kf, PACKAGE_DATA_DIR "/archivers.list", 0, NULL))
     {
-        int n_archivers;
-        char** programs = g_key_file_get_groups(kf, &n_archivers);
+        gsize n_archivers;
+        gchar** programs = g_key_file_get_groups(kf, &n_archivers);
         if(programs)
         {
-            int i;
+            gsize i;
             for(i = 0; i < n_archivers; ++i)
             {
                 FmArchiver* archiver = g_slice_new0(FmArchiver);
