@@ -19,6 +19,17 @@
  *      MA 02110-1301, USA.
  */
 
+/**
+ * SECTION:fm-bookmarks
+ * @short_description: Bookmarks support for libfm.
+ * @title: FmBookmarks
+ *
+ * @include: libfm/fm-bookmarks.h
+ *
+ * The application that uses libfm can use user-wide bookmark list via
+ * class FmBookmarks.
+ */
+
 #include "fm-bookmarks.h"
 #include <stdio.h>
 #include <string.h>
@@ -49,6 +60,15 @@ static void fm_bookmarks_class_init(FmBookmarksClass *klass)
     g_object_class = G_OBJECT_CLASS(klass);
     g_object_class->finalize = fm_bookmarks_finalize;
 
+    /**
+     * FmBookmarks::changed:
+     * @bookmarks: pointer to bookmarks list singleton descriptor
+     *
+     * The "changed" signal is emitted when some bookmark item is
+     * changed, added, or removed.
+     *
+     * Since: 0.1.0
+     */
     signals[CHANGED] =
         g_signal_new("changed",
                      G_TYPE_FROM_CLASS(klass),
@@ -64,7 +84,7 @@ static void fm_bookmarks_finalize(GObject *object)
     FmBookmarks *self;
 
     g_return_if_fail(object != NULL);
-    g_return_if_fail(IS_FM_BOOKMARKS(object));
+    g_return_if_fail(FM_IS_BOOKMARKS(object));
 
     self = FM_BOOKMARKS(object);
 
@@ -82,12 +102,12 @@ static void fm_bookmarks_finalize(GObject *object)
     G_OBJECT_CLASS(fm_bookmarks_parent_class)->finalize(object);
 }
 
-char* get_bookmarks_file()
+static char* get_bookmarks_file()
 {
     return g_build_filename(g_get_home_dir(), ".gtk-bookmarks", NULL);
 }
 
-void free_item(FmBookmarkItem* item)
+static void free_item(FmBookmarkItem* item)
 {
     if(item->name != item->path->name)
         g_free(item->name);
@@ -129,7 +149,7 @@ static FmBookmarkItem* new_item(char* line)
     return item;
 }
 
-GList* load_bookmarks(const char* fpath)
+static GList* load_bookmarks(const char* fpath)
 {
     FILE* f;
     char buf[1024];
@@ -137,13 +157,15 @@ GList* load_bookmarks(const char* fpath)
     GList* items = NULL;
 
     /* load the file */
-    if( f = fopen(fpath, "r") )
+    f = fopen(fpath, "r");
+    if(f)
     {
         while(fgets(buf, 1024, f))
         {
             item = new_item(buf);
             items = g_list_prepend(items, item);
         }
+        fclose(f);
     }
     items = g_list_reverse(items);
     return items;
@@ -161,29 +183,49 @@ static void fm_bookmarks_init(FmBookmarks *self)
     g_free(fpath);
 }
 
-FmBookmarks *fm_bookmarks_new(void)
+static FmBookmarks *fm_bookmarks_new(void)
 {
     return g_object_new(FM_BOOKMARKS_TYPE, NULL);
 }
 
-FmBookmarks* fm_bookmarks_get(void)
+/**
+ * fm_bookmarks_dup
+ *
+ * Returns reference to bookmarks list singleton descriptor.
+ *
+ * Return value: (transfer full): a reference to bookmarks list
+ *
+ * Since: 0.1.99
+ */
+FmBookmarks* fm_bookmarks_dup(void)
 {
     if( G_LIKELY(singleton) )
         g_object_ref(singleton);
     else
     {
         singleton = fm_bookmarks_new();
-        g_object_add_weak_pointer(G_OBJECT(singleton), &singleton);
+        g_object_add_weak_pointer(G_OBJECT(singleton), (gpointer*)&singleton);
     }
     return singleton;
 }
 
-GList* fm_bookmarks_list_all(FmBookmarks* bookmarks)
+/**
+ * fm_bookmarks_list_all
+ * @bookmarks: bookmarks list
+ *
+ * Returns list of FmBookmarkItem retrieved from bookmarks list. Returned
+ * list is owned by bookmarks list and should not be freed by caller.
+ *
+ * Return value: (transfer none): (element-type #FmBookmarkItem): list of bookmark items
+ *
+ * Since: 0.1.0
+ */
+const GList* fm_bookmarks_list_all(FmBookmarks* bookmarks)
 {
     return bookmarks->items;
 }
 
-gboolean save_bookmarks(FmBookmarks* bookmarks)
+static gboolean save_bookmarks(FmBookmarks* bookmarks)
 {
     FmBookmarkItem* item;
     GList* l;
@@ -210,13 +252,27 @@ gboolean save_bookmarks(FmBookmarks* bookmarks)
     return FALSE;
 }
 
-void queue_save_bookmarks(FmBookmarks* bookmarks)
+static void queue_save_bookmarks(FmBookmarks* bookmarks)
 {
     if(idle_handler)
         g_source_remove(idle_handler);
     idle_handler = g_idle_add((GSourceFunc)save_bookmarks, bookmarks);
 }
 
+/**
+ * fm_bookmarks_insert
+ * @bookmarks: bookmarks list
+ * @path: path requested to add to bookmarks
+ * @name: name new bookmark will be seen in list with
+ * @pos: where to insert a bookmark into list
+ *
+ * Adds a bookmark into bookmark list. Returned structure is managed by
+ * bookmarks list and should not be freed by caller.
+ *
+ * Return value: (transfer none): new created bookmark item
+ *
+ * Since: 0.1.0
+ */
 FmBookmarkItem* fm_bookmarks_insert(FmBookmarks* bookmarks, FmPath* path, const char* name, int pos)
 {
     FmBookmarkItem* item = g_slice_new0(FmBookmarkItem);
@@ -228,6 +284,15 @@ FmBookmarkItem* fm_bookmarks_insert(FmBookmarks* bookmarks, FmPath* path, const
     return item;
 }
 
+/**
+ * fm_bookmarks_remove
+ * @bookmarks: bookmarks list
+ * @item: bookmark item for deletion
+ *
+ * Removes a bookmark from bookmark list.
+ *
+ * Since: 0.1.0
+ */
 void fm_bookmarks_remove(FmBookmarks* bookmarks, FmBookmarkItem* item)
 {
     bookmarks->items = g_list_remove(bookmarks->items, item);
@@ -235,6 +300,16 @@ void fm_bookmarks_remove(FmBookmarks* bookmarks, FmBookmarkItem* item)
     queue_save_bookmarks(bookmarks);
 }
 
+/**
+ * fm_bookmarks_rename
+ * @bookmarks: bookmarks list
+ * @item: bookmark item which will be changed
+ * @new_name: new name for bookmark item to be seen in list
+ *
+ * Changes name of existing bookmark item.
+ *
+ * Since: 0.1.0
+ */
 void fm_bookmarks_rename(FmBookmarks* bookmarks, FmBookmarkItem* item, const char* new_name)
 {
     g_free(item->name);
@@ -242,9 +317,18 @@ void fm_bookmarks_rename(FmBookmarks* bookmarks, FmBookmarkItem* item, const cha
     queue_save_bookmarks(bookmarks);
 }
 
+/**
+ * fm_bookmarks_reorder
+ * @bookmarks: bookmarks list
+ * @item: bookmark item which will be changed
+ * @pos: new position for bookmark item in list
+ *
+ * Changes position of existing bookmark item.
+ *
+ * Since: 0.1.0
+ */
 void fm_bookmarks_reorder(FmBookmarks* bookmarks, FmBookmarkItem* item, int pos)
 {
-    // GList* l = g_list_find(biikmarks->items, item);
     bookmarks->items = g_list_remove(bookmarks->items, item);
     bookmarks->items = g_list_insert(bookmarks->items, item, pos);
     queue_save_bookmarks(bookmarks);
