@@ -30,13 +30,22 @@ G_BEGIN_DECLS
 
 typedef gboolean (*FmLaunchFolderFunc)(GAppLaunchContext* ctx, GList* folder_infos, gpointer user_data, GError** err);
 
+typedef enum {
+    FM_FILE_LAUNCHER_EXEC = 1,
+    FM_FILE_LAUNCHER_EXEC_IN_TERMINAL,
+    FM_FILE_LAUNCHER_EXEC_OPEN,
+    FM_FILE_LAUNCHER_EXEC_CANCEL
+} FmFileLauncherExecAction;
+
 typedef struct _FmFileLauncher FmFileLauncher;
 struct _FmFileLauncher
 {
     GAppInfo* (*get_app)(GList* file_infos, FmMimeType* mime_type, gpointer user_data, GError** err);
     /* gboolean (*before_open)(GAppLaunchContext* ctx, GList* folder_infos, gpointer user_data); */
     gboolean (*open_folder)(GAppLaunchContext* ctx, GList* folder_infos, gpointer user_data, GError** err);
+    FmFileLauncherExecAction (*exec_file)(FmFileInfo* file, gpointer user_data);
     gboolean (*error)(GAppLaunchContext* ctx, GError* err, gpointer user_data);
+    int (*ask)(const char* msg, char* const* btn_labels, int default_btn, gpointer user_data);
 };
 
 gboolean fm_launch_files(GAppLaunchContext* ctx, GList* file_infos, FmFileLauncher* launcher, gpointer user_data);
