@@ -28,13 +28,28 @@
 
 G_BEGIN_DECLS
 
-char* fm_file_size_to_str( char* buf, goffset size, gboolean si_prefix );
+typedef const char* (*FmAppCommandParseCallback)(char opt, gpointer user_data);
+
+typedef struct _FmAppCommandParseOption FmAppCommandParseOption;
+
+struct _FmAppCommandParseOption
+{
+    char opt;
+    FmAppCommandParseCallback callback;
+};
+
+int fm_app_command_parse(const char* cmd, const FmAppCommandParseOption* opts,
+                         char** ret, gpointer user_data);
+
+char* fm_file_size_to_str(char* buf, size_t buf_size, goffset size, gboolean si_prefix);
 
 gboolean fm_key_file_get_int(GKeyFile* kf, const char* grp, const char* key, int* val);
 gboolean fm_key_file_get_bool(GKeyFile* kf, const char* grp, const char* key, gboolean* val);
 
 char* fm_canonicalize_filename(const char* filename, const char* cwd);
 
+char* fm_strdup_replace(char* str, char* old_str, char* new_str);
+
 G_END_DECLS
 
 #endif
