@@ -38,41 +38,53 @@ G_BEGIN_DECLS
 #define FM_IS_DND_DEST_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass),\
             FM_TYPE_DND_DEST))
 
-/* default droppable targets */
-enum
+/**
+ * FmDndDestTargetType
+ * @FM_DND_DEST_TARGET_FM_LIST: direct pointer of #FmList
+ * @FM_DND_DEST_TARGET_URI_LIST: "text/uri-list"
+ * @FM_DND_DEST_TARGET_XDS: X direct save
+ *
+ * default droppable targets supported by #FmDndDest
+ */
+typedef enum
 {
-    FM_DND_DEST_TARGET_FM_LIST, /* direct pointer of FmList */
-    FM_DND_DEST_TARGET_URI_LIST, /* text/uri-list */
-    FM_DND_DEST_TARGET_XDS, /* X direct save */
+    FM_DND_DEST_TARGET_FM_LIST,
+    FM_DND_DEST_TARGET_URI_LIST,
+    FM_DND_DEST_TARGET_XDS,
+    /*< private >*/
     N_FM_DND_DEST_DEFAULT_TARGETS
-};
-
-extern GtkTargetEntry fm_default_dnd_dest_targets[];
+} FmDndDestTargetType;
 
 typedef struct _FmDndDest           FmDndDest;
 typedef struct _FmDndDestClass      FmDndDestClass;
 
+/**
+ * FmDndDestClass
+ * @parent_class: the parent class
+ * @files_dropped: the class closure for the #FmDndDest::files-dropped signal
+ */
 struct _FmDndDestClass
 {
     GObjectClass parent_class;
-    gboolean (*files_dropped)(FmDndDest* dd, int x, int y, guint action, guint info_type, FmFileInfoList* files);
+    gboolean (*files_dropped)(FmDndDest* dd, int x, int y, guint action, guint info_type, FmPathList* files);
 };
 
 GType       fm_dnd_dest_get_type        (void);
 FmDndDest*  fm_dnd_dest_new         (GtkWidget* w);
+FmDndDest*  fm_dnd_dest_new_with_handlers (GtkWidget* w);
 
 void fm_dnd_dest_set_widget(FmDndDest* dd, GtkWidget* w);
 
-/* the returned list can be either FmPathList or FmFileInfoList */
-/* check with fm_list_is_path_list() and fm_list_is_file_info_list(). */
-FmList* fm_dnd_dest_get_src_files(FmDndDest* dd);
-
 void fm_dnd_dest_set_dest_file(FmDndDest* dd, FmFileInfo* dest_file);
 FmFileInfo* fm_dnd_dest_get_dest_file(FmDndDest* dd);
 FmPath* fm_dnd_dest_get_dest_path(FmDndDest* dd);
 
+#if !GTK_CHECK_VERSION(2, 22, 0)
+#  define gdk_drag_context_list_targets(ctx) ctx->targets
+#endif
+
 #define fm_drag_context_has_target(ctx, target) \
-    (g_list_find(ctx->targets, target) != NULL)
+    (g_list_find(gdk_drag_context_list_targets(ctx), target) != NULL)
 
 #define fm_drag_context_has_target_name(ctx, name)  \
     fm_drag_context_has_target(ctx, gdk_atom_intern_static_string(name))
@@ -89,10 +101,24 @@ gboolean fm_dnd_dest_drag_drop(FmDndDest* dd, GdkDragContext *drag_context,
 
 GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
                                              GdkDragContext* drag_context,
-                                             GdkTarget target);
+                                             GdkAtom target);
 
 void fm_dnd_dest_drag_leave(FmDndDest* dd, GdkDragContext* drag_context, guint time);
 
+/**
+ * fm_dnd_dest_add_targets
+ * @widget: #GtkWidget to add targets
+ * @targets: pointer to array of #GtkTargetEntry to add
+ * @n: number of targets to add
+ *
+ * Adds drop destination targets to existing list for @widget. Convenience API.
+ *
+ * Since: 1.0.1
+ */
+#define fm_dnd_dest_add_targets(widget,targets,n) \
+            gtk_target_list_add_table(gtk_drag_dest_get_target_list(widget), \
+                                      targets, n)
+
 G_END_DECLS
 
 #endif /* __FM_DND_DEST_H__ */
