@@ -2,6 +2,7 @@
  *      fm-gtk-utils.c
  *
  *      Copyright 2009 PCMan <pcman@debian>
+ *      Copyright 2012 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -22,6 +23,7 @@
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
+#include "gtk-compat.h"
 
 #include <glib/gi18n-lib.h>
 #include <gio/gdesktopappinfo.h>
@@ -37,73 +39,117 @@
 static GtkDialog*   _fm_get_user_input_dialog   (GtkWindow* parent, const char* title, const char* msg);
 static gchar*       _fm_user_input_dialog_run   (GtkDialog* dlg, GtkEntry *entry);
 
-void fm_show_error(GtkWindow* parent, const char* msg)
+void fm_show_error(GtkWindow* parent, const char* title, const char* msg)
 {
     GtkWidget* dlg = gtk_message_dialog_new(parent, 0,
                                             GTK_MESSAGE_ERROR,
-                                            GTK_BUTTONS_OK, msg);
-    gtk_window_set_title((GtkWindow*)dlg, _("Error"));
-    gtk_dialog_run((GtkDialog*)dlg);
+                                            GTK_BUTTONS_OK, "%s", msg);
+    gtk_window_set_title(GTK_WINDOW(dlg), title ? title : _("Error"));
+    gtk_dialog_run(GTK_DIALOG(dlg));
     gtk_widget_destroy(dlg);
 }
 
-gboolean fm_yes_no(GtkWindow* parent, const char* question, gboolean default_yes)
+gboolean fm_yes_no(GtkWindow* parent, const char* title, const char* question, gboolean default_yes)
 {
     int ret;
-    GtkWidget* dlg = gtk_message_dialog_new_with_markup(parent, 0,
-                                GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, question);
-    gtk_dialog_set_default_response(GTK_DIALOG(dlg), default_yes ? GTK_RESPONSE_YES : GTK_RESPONSE_NO);
-    ret = gtk_dialog_run((GtkDialog*)dlg);
-    gtk_widget_destroy(dlg);
+    GtkDialog* dlg = GTK_DIALOG(gtk_message_dialog_new_with_markup(parent, 0,
+                                                        GTK_MESSAGE_QUESTION,
+                                                        GTK_BUTTONS_YES_NO,
+                                                        "%s", question));
+    gtk_window_set_title(GTK_WINDOW(dlg), title ? title : _("Confirm"));
+    gtk_dialog_set_default_response(dlg, default_yes ? GTK_RESPONSE_YES : GTK_RESPONSE_NO);
+    /* #3300797: Delete prompt isn't on the first layer */
+    gtk_window_set_keep_above(GTK_WINDOW(dlg), TRUE);
+    ret = gtk_dialog_run(dlg);
+    gtk_widget_destroy((GtkWidget*)dlg);
     return ret == GTK_RESPONSE_YES;
 }
 
-gboolean fm_ok_cancel(GtkWindow* parent, const char* question, gboolean default_ok)
+gboolean fm_ok_cancel(GtkWindow* parent, const char* title, const char* question, gboolean default_ok)
 {
     int ret;
-    GtkWidget* dlg = gtk_message_dialog_new_with_markup(parent, 0,
-                                GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL, question);
-    gtk_dialog_set_default_response(GTK_DIALOG(dlg), default_ok ? GTK_RESPONSE_OK : GTK_RESPONSE_CANCEL);
-    ret = gtk_dialog_run((GtkDialog*)dlg);
-    gtk_widget_destroy(dlg);
+    GtkDialog* dlg = GTK_DIALOG(gtk_message_dialog_new_with_markup(parent, 0,
+                                                        GTK_MESSAGE_QUESTION,
+                                                        GTK_BUTTONS_OK_CANCEL,
+                                                        "%s", question));
+    gtk_window_set_title(GTK_WINDOW(dlg), title ? title : _("Confirm"));
+    gtk_dialog_set_default_response(dlg, default_ok ? GTK_RESPONSE_OK : GTK_RESPONSE_CANCEL);
+    ret = gtk_dialog_run(dlg);
+    gtk_widget_destroy((GtkWidget*)dlg);
     return ret == GTK_RESPONSE_OK;
 }
 
-int fm_ask(GtkWindow* parent, const char* question, ...)
+/**
+ * fm_ask
+ * @parent: toplevel parent widget
+ * @title: title for the window with question
+ * @question: the question to show to the user
+ * @...: a NULL terminated list of button labels
+ *
+ * Ask the user a question with several options provided.
+ *
+ * Return value: the index of selected button, or -1 if the dialog is closed.
+ */
+int fm_ask(GtkWindow* parent, const char* title, const char* question, ...)
 {
     int ret;
     va_list args;
     va_start (args, question);
-    ret = fm_ask_valist(parent, question, args);
+    ret = fm_ask_valist(parent, title, question, args);
     va_end (args);
     return ret;
 }
 
-int fm_askv(GtkWindow* parent, const char* question, const char** options)
+/**
+ * fm_askv
+ * @parent: toplevel parent widget
+ * @title: title for the window with question
+ * @question: the question to show to the user
+ * @options: a NULL terminated list of button labels
+ *
+ * Ask the user a question with several options provided.
+ *
+ * Return value: the index of selected button, or -1 if the dialog is closed.
+ */
+int fm_askv(GtkWindow* parent, const char* title, const char* question, char* const* options)
 {
     int ret;
     guint id = 1;
-    GtkWidget* dlg = gtk_message_dialog_new_with_markup(parent, 0,
-                                GTK_MESSAGE_QUESTION, 0, question);
+    GtkDialog* dlg = GTK_DIALOG(gtk_message_dialog_new_with_markup(parent, 0,
+                                                        GTK_MESSAGE_QUESTION, 0,
+                                                        "%s", question));
+    gtk_window_set_title(GTK_WINDOW(dlg), title ? title : _("Question"));
     /* FIXME: need to handle defualt button and alternative button
      * order problems. */
     while(*options)
     {
         /* FIXME: handle button image and stock buttons */
-        GtkWidget* btn = gtk_dialog_add_button(GTK_DIALOG( dlg ), *options, id);
+        /*GtkWidget* btn =*/
+        gtk_dialog_add_button(dlg, *options, id);
         ++options;
         ++id;
     }
-    ret = gtk_dialog_run((GtkDialog*)dlg);
+    ret = gtk_dialog_run(dlg);
     if(ret >= 1)
         ret -= 1;
     else
-        ret == -1;
-    gtk_widget_destroy(dlg);
+        ret = -1;
+    gtk_widget_destroy((GtkWidget*)dlg);
     return ret;
 }
 
-int fm_ask_valist(GtkWindow* parent, const char* question, va_list options)
+/**
+ * fm_ask_valist
+ * @parent: toplevel parent widget
+ * @title: title for the window with question
+ * @question: the question to show to the user
+ * @options: va_arg list of button labels
+ *
+ * Ask the user a question with several options provided.
+ *
+ * Return value: the index of selected button, or -1 if the dialog is closed.
+ */
+int fm_ask_valist(GtkWindow* parent, const char* title, const char* question, va_list options)
 {
     GArray* opts = g_array_sized_new(TRUE, TRUE, sizeof(char*), 6);
     gint ret;
@@ -113,7 +159,7 @@ int fm_ask_valist(GtkWindow* parent, const char* question, va_list options)
         g_array_append_val(opts, opt);
         opt = va_arg (options, const char *);
     }
-    ret = fm_askv(parent, question, opts->data);
+    ret = fm_askv(parent, title, question, &opts->data);
     g_array_free(opts, TRUE);
     return ret;
 }
@@ -123,33 +169,33 @@ int fm_ask_valist(GtkWindow* parent, const char* question, va_list options)
 gchar* fm_get_user_input(GtkWindow* parent, const char* title, const char* msg, const char* default_text)
 {
     GtkDialog* dlg = _fm_get_user_input_dialog( parent, title, msg);
-    GtkWidget* entry = gtk_entry_new();
-    gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
+    GtkEntry* entry = GTK_ENTRY(gtk_entry_new());
+    gtk_entry_set_activates_default(entry, TRUE);
 
     if(default_text && default_text[0])
-        gtk_entry_set_text(GTK_ENTRY( entry ), default_text);
+        gtk_entry_set_text(entry, default_text);
 
-    return _fm_user_input_dialog_run( dlg,  GTK_ENTRY( entry ) );
+    return _fm_user_input_dialog_run(dlg, entry);
 }
 
 FmPath* fm_get_user_input_path(GtkWindow* parent, const char* title, const char* msg, FmPath* default_path)
 {
 
     GtkDialog* dlg = _fm_get_user_input_dialog( parent, title, msg);
-    GtkWidget* entry = gtk_entry_new();
+    GtkEntry* entry = GTK_ENTRY(gtk_entry_new());
     char *str, *path_str = NULL;
     FmPath* path;
 
-    gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
+    gtk_entry_set_activates_default(entry, TRUE);
 
     if(default_path)
     {
         path_str = fm_path_display_name(default_path, FALSE);
-        gtk_entry_set_text(GTK_ENTRY( entry ), path_str);
+        gtk_entry_set_text(entry, path_str);
     }
 
-    str = _fm_user_input_dialog_run( dlg,  GTK_ENTRY( entry ) );
-    path = fm_path_new(str);
+    str = _fm_user_input_dialog_run(dlg, entry);
+    path = fm_path_new_for_str(str);
 
     g_free(path_str);
     g_free(str);
@@ -160,12 +206,12 @@ FmPath* fm_get_user_input_path(GtkWindow* parent, const char* title, const char*
 gchar* fm_get_user_input_rename(GtkWindow* parent, const char* title, const char* msg, const char* default_text)
 {
     GtkDialog* dlg = _fm_get_user_input_dialog( parent, title, msg);
-    GtkWidget* entry = gtk_entry_new();
-    gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
+    GtkEntry* entry = GTK_ENTRY(gtk_entry_new());
+    gtk_entry_set_activates_default(entry, TRUE);
 
     if(default_text && default_text[0])
     {
-        gtk_entry_set_text(GTK_ENTRY( entry ), default_text);
+        gtk_entry_set_text(entry, default_text);
         /* only select filename part without extension name. */
         if(default_text[1])
         {
@@ -196,30 +242,35 @@ gchar* fm_get_user_input_rename(GtkWindow* parent, const char* title, const char
         }
     }
 
-    return _fm_user_input_dialog_run( dlg,  GTK_ENTRY( entry ) );
+    return _fm_user_input_dialog_run(dlg, entry);
 }
 
 static GtkDialog* _fm_get_user_input_dialog(GtkWindow* parent, const char* title, const char* msg)
 {
-    GtkWidget* dlg = gtk_dialog_new_with_buttons(title, parent, GTK_DIALOG_NO_SEPARATOR,
+    GtkDialog* dlg = GTK_DIALOG(gtk_dialog_new_with_buttons(title, parent,
+#if GTK_CHECK_VERSION(3, 0, 0)
+                                0,
+#else
+                                GTK_DIALOG_NO_SEPARATOR,
+#endif
                                 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-                                GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);
+                                GTK_STOCK_OK, GTK_RESPONSE_OK, NULL));
     GtkWidget* label = gtk_label_new(msg);
     gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
 
-    gtk_dialog_set_alternative_button_order(GTK_DIALOG(dlg), GTK_RESPONSE_OK, GTK_RESPONSE_CANCEL, -1);
-    gtk_box_set_spacing((GtkBox*)gtk_dialog_get_content_area(GTK_DIALOG(dlg)), 6);
-    gtk_box_pack_start((GtkBox*)gtk_dialog_get_content_area(GTK_DIALOG(dlg)), label, FALSE, TRUE, 6);
+    gtk_dialog_set_alternative_button_order(dlg, GTK_RESPONSE_OK, GTK_RESPONSE_CANCEL, -1);
+    gtk_box_set_spacing((GtkBox*)gtk_dialog_get_content_area(dlg), 6);
+    gtk_box_pack_start((GtkBox*)gtk_dialog_get_content_area(dlg), label, FALSE, TRUE, 6);
 
-    gtk_container_set_border_width(GTK_CONTAINER((GtkBox*)gtk_dialog_get_content_area(GTK_DIALOG(dlg))), 12);
+    gtk_container_set_border_width(GTK_CONTAINER((GtkBox*)gtk_dialog_get_content_area(dlg)), 12);
     gtk_container_set_border_width(GTK_CONTAINER(dlg), 5);
-    gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_OK);
+    gtk_dialog_set_default_response(dlg, GTK_RESPONSE_OK);
     gtk_window_set_default_size(GTK_WINDOW(dlg), 480, -1);
 
     return dlg;
 }
 
-static gchar* _fm_user_input_dialog_run( GtkDialog* dlg, GtkEntry *entry)
+static gchar* _fm_user_input_dialog_run(GtkDialog* dlg, GtkEntry *entry)
 {
     char* str = NULL;
     int sel_start, sel_end;
@@ -228,7 +279,8 @@ static gchar* _fm_user_input_dialog_run( GtkDialog* dlg, GtkEntry *entry)
     /* FIXME: this workaround is used to overcome bug of gtk+.
      * gtk+ seems to ignore select region and select all text for entry in dialog. */
     has_sel = gtk_editable_get_selection_bounds(GTK_EDITABLE(entry), &sel_start, &sel_end);
-    gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dlg)->vbox ), GTK_WIDGET( entry ), FALSE, TRUE, 6);
+    gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(dlg)),
+                       GTK_WIDGET(entry), FALSE, TRUE, 6);
     gtk_widget_show_all(GTK_WIDGET(dlg));
 
     if(has_sel)
@@ -247,29 +299,111 @@ static gchar* _fm_user_input_dialog_run( GtkDialog* dlg, GtkEntry *entry)
     return str;
 }
 
-FmPath* fm_select_folder(GtkWindow* parent)
+static void on_update_img_preview( GtkFileChooser *chooser, GtkImage* img )
+{
+    char* file = gtk_file_chooser_get_preview_filename(chooser);
+    GdkPixbuf* pix = NULL;
+    if(file)
+    {
+        pix = gdk_pixbuf_new_from_file_at_scale( file, 128, 128, TRUE, NULL );
+        g_free( file );
+    }
+    if(pix)
+    {
+        gtk_file_chooser_set_preview_widget_active(chooser, TRUE);
+        gtk_image_set_from_pixbuf(img, pix);
+        g_object_unref(pix);
+    }
+    else
+    {
+        gtk_image_clear(img);
+        gtk_file_chooser_set_preview_widget_active(chooser, FALSE);
+    }
+}
+
+static gulong fm_add_image_preview_to_file_chooser(GtkFileChooser* chooser)
+{
+    GtkWidget* img_preview = gtk_image_new();
+    gtk_misc_set_alignment(GTK_MISC(img_preview), 0.5, 0.0);
+    gtk_widget_set_size_request(img_preview, 128, 128);
+    gtk_file_chooser_set_preview_widget(chooser, img_preview);
+    return g_signal_connect(chooser, "update-preview", G_CALLBACK(on_update_img_preview), img_preview);
+}
+
+/* TODO: support selecting multiple files */
+FmPath* fm_select_file(GtkWindow* parent, 
+                        const char* title, 
+                        const char* default_folder,
+                        gboolean local_only,
+                        gboolean show_preview,
+                        /* filter1, filter2, ..., NULL */ ...)
 {
     FmPath* path;
     GtkFileChooser* chooser;
-    chooser = (GtkFileChooser*)gtk_file_chooser_dialog_new(_("Please select a folder"),
+    GtkFileFilter* filter;
+    gulong handler_id = 0;
+    va_list args;
+
+    chooser = (GtkFileChooser*)gtk_file_chooser_dialog_new(
+                                        title, parent, GTK_FILE_CHOOSER_ACTION_OPEN,
+                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+                                        GTK_STOCK_OK, GTK_RESPONSE_OK,
+                                        NULL);
+    gtk_dialog_set_alternative_button_order(GTK_DIALOG(chooser),
+                                        GTK_RESPONSE_CANCEL,
+                                        GTK_RESPONSE_OK, NULL);
+    if(local_only)
+        gtk_file_chooser_set_local_only(chooser, TRUE);
+
+    if(default_folder)
+        gtk_file_chooser_set_current_folder(chooser, default_folder);
+
+    va_start(args, show_preview);
+    while((filter = va_arg(args, GtkFileFilter*)))
+    {
+        gtk_file_chooser_add_filter(chooser, filter);
+    }
+    va_end (args);
+
+    if(show_preview)
+        handler_id = fm_add_image_preview_to_file_chooser(chooser);
+
+    if(gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_OK)
+    {
+        GFile* file = gtk_file_chooser_get_file(chooser);
+        path = fm_path_new_for_gfile(file);
+        g_object_unref(file);
+    }
+    else
+        path = NULL;
+    if(handler_id > 0)
+        g_signal_handler_disconnect(chooser, handler_id);
+    gtk_widget_destroy(GTK_WIDGET(chooser));
+    return path;
+}
+
+FmPath* fm_select_folder(GtkWindow* parent, const char* title)
+{
+    FmPath* path;
+    GtkFileChooser* chooser;
+    chooser = (GtkFileChooser*)gtk_file_chooser_dialog_new(
+                                        title ? title : _("Please select a folder"),
                                         parent, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
                                         NULL);
-    gtk_dialog_set_alternative_button_order((GtkDialog*)chooser,
+    gtk_dialog_set_alternative_button_order(GTK_DIALOG(chooser),
                                         GTK_RESPONSE_CANCEL,
                                         GTK_RESPONSE_OK, NULL);
-    if( gtk_dialog_run((GtkDialog*)chooser) == GTK_RESPONSE_OK )
+    if( gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_OK )
     {
-        char* file = gtk_file_chooser_get_filename(chooser);
-        if(!file)
-            file = gtk_file_chooser_get_uri(chooser);
-        path = fm_path_new(file);
-        g_free(file);
+        GFile* file = gtk_file_chooser_get_file(chooser);
+        path = fm_path_new_for_gfile(file);
+        g_object_unref(file);
     }
     else
         path = NULL;
-    gtk_widget_destroy((GtkWidget*)chooser);
+    gtk_widget_destroy(GTK_WIDGET(chooser));
     return path;
 }
 
@@ -293,7 +427,7 @@ struct MountData
 static void on_mount_action_finished(GObject* src, GAsyncResult *res, gpointer user_data)
 {
     struct MountData* data = user_data;
-g_debug("on_mount_action_finished");
+
     switch(data->action)
     {
     case MOUNT_VOLUME:
@@ -303,30 +437,34 @@ g_debug("on_mount_action_finished");
         data->ret = g_file_mount_enclosing_volume_finish(G_FILE(src), res, &data->err);
         break;
     case UMOUNT_MOUNT:
-#if GLIB_CHECK_VERSION(2, 22, 0)
         data->ret = g_mount_unmount_with_operation_finish(G_MOUNT(src), res, &data->err);
-#else
-        data->ret = g_mount_unmount_finish(G_MOUNT(src), res, &data->err);
-#endif
         break;
     case EJECT_MOUNT:
-#if GLIB_CHECK_VERSION(2, 22, 0)
         data->ret = g_mount_eject_with_operation_finish(G_MOUNT(src), res, &data->err);
-#else
-        data->ret = g_mount_eject_finish(G_MOUNT(src), res, &data->err);
-#endif
         break;
     case EJECT_VOLUME:
-#if GLIB_CHECK_VERSION(2, 22, 0)
         data->ret = g_volume_eject_with_operation_finish(G_VOLUME(src), res, &data->err);
-#else
-        data->ret = g_volume_eject_finish(G_VOLUME(src), res, &data->err);
-#endif
         break;
     }
     g_main_loop_quit(data->loop);
 }
 
+static void prepare_unmount(GMount* mount)
+{
+    /* ensure that CWD is not on the mounted filesystem. */
+    char* cwd_str = g_get_current_dir();
+    GFile* cwd = g_file_new_for_path(cwd_str);
+    GFile* root = g_mount_get_root(mount);
+    g_free(cwd_str);
+    /* FIXME: This cannot cover 100% cases since symlinks are not checked.
+     * There may be other cases that cwd is actually under mount root
+     * but checking prefix is not enough. We already did our best, though. */
+    if(g_file_has_prefix(cwd, root))
+        g_chdir("/");
+    g_object_unref(cwd);
+    g_object_unref(root);
+}
+
 static gboolean fm_do_mount(GtkWindow* parent, GObject* obj, MountAction action, gboolean interactive)
 {
     gboolean ret;
@@ -346,6 +484,7 @@ static gboolean fm_do_mount(GtkWindow* parent, GObject* obj, MountAction action,
         g_file_mount_enclosing_volume(G_FILE(obj), 0, op, cancellable, on_mount_action_finished, data);
         break;
     case UMOUNT_MOUNT:
+        prepare_unmount(G_MOUNT(obj));
 #if GLIB_CHECK_VERSION(2, 22, 0)
         g_mount_unmount_with_operation(G_MOUNT(obj), G_MOUNT_UNMOUNT_NONE, op, cancellable, on_mount_action_finished, data);
 #else
@@ -353,6 +492,7 @@ static gboolean fm_do_mount(GtkWindow* parent, GObject* obj, MountAction action,
 #endif
         break;
     case EJECT_MOUNT:
+        prepare_unmount(G_MOUNT(obj));
 #if GLIB_CHECK_VERSION(2, 22, 0)
         g_mount_eject_with_operation(G_MOUNT(obj), G_MOUNT_UNMOUNT_NONE, op, cancellable, on_mount_action_finished, data);
 #else
@@ -360,11 +500,16 @@ static gboolean fm_do_mount(GtkWindow* parent, GObject* obj, MountAction action,
 #endif
         break;
     case EJECT_VOLUME:
+        {
+            GMount* mnt = g_volume_get_mount(G_VOLUME(obj));
+            prepare_unmount(mnt);
+            g_object_unref(mnt);
 #if GLIB_CHECK_VERSION(2, 22, 0)
-        g_volume_eject_with_operation(G_VOLUME(obj), G_MOUNT_UNMOUNT_NONE, op, cancellable, on_mount_action_finished, data);
+            g_volume_eject_with_operation(G_VOLUME(obj), G_MOUNT_UNMOUNT_NONE, op, cancellable, on_mount_action_finished, data);
 #else
-        g_volume_eject(G_VOLUME(obj), G_MOUNT_UNMOUNT_NONE, cancellable, on_mount_action_finished, data);
+            g_volume_eject(G_VOLUME(obj), G_MOUNT_UNMOUNT_NONE, cancellable, on_mount_action_finished, data);
 #endif
+        }
         break;
     }
 
@@ -407,7 +552,7 @@ static gboolean fm_do_mount(GtkWindow* parent, GObject* obj, MountAction action,
                     interactive = FALSE;
             }
             if(interactive)
-                fm_show_error(parent, data->err->message);
+                fm_show_error(parent, NULL, data->err->message);
         }
         g_error_free(data->err);
     }
@@ -463,55 +608,71 @@ gboolean fm_eject_volume(GtkWindow* parent, GVolume* vol, gboolean interactive)
 /* FIXME: only show the progress dialog if the job isn't finished
  * in 1 sec. */
 
-void fm_copy_files(FmPathList* files, FmPath* dest_dir)
+void fm_copy_files(GtkWindow* parent, FmPathList* files, FmPath* dest_dir)
 {
-    FmJob* job = fm_file_ops_job_new(FM_FILE_OP_COPY, files);
-    fm_file_ops_job_set_dest(FM_FILE_OPS_JOB(job), dest_dir);
-    fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
+    FmFileOpsJob* job = fm_file_ops_job_new(FM_FILE_OP_COPY, files);
+    fm_file_ops_job_set_dest(job, dest_dir);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
 }
 
-void fm_move_files(FmPathList* files, FmPath* dest_dir)
+void fm_move_files(GtkWindow* parent, FmPathList* files, FmPath* dest_dir)
 {
-    FmJob* job = fm_file_ops_job_new(FM_FILE_OP_MOVE, files);
-    fm_file_ops_job_set_dest(FM_FILE_OPS_JOB(job), dest_dir);
-    fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
+    FmFileOpsJob* job = fm_file_ops_job_new(FM_FILE_OP_MOVE, files);
+    fm_file_ops_job_set_dest(job, dest_dir);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
 }
 
-void fm_trash_files(FmPathList* files)
+/**
+ * fm_link_files
+ * @parent:   window to base progress dialog over it
+ * @files:    list of files to make symbolic links to
+ * @dest_dir: directory where symbolic links should be created
+ *
+ * Create symbolic links for some files in the target directory with
+ * progress dialog.
+ */
+void fm_link_files(GtkWindow* parent, FmPathList* files, FmPath* dest_dir)
 {
-    if(!fm_config->confirm_del || fm_yes_no(NULL, _("Do you want to move the selected files to trash can?"), TRUE))
+    FmFileOpsJob* job = fm_file_ops_job_new(FM_FILE_OP_LINK, files);
+    fm_file_ops_job_set_dest(job, dest_dir);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
+}
+
+void fm_trash_files(GtkWindow* parent, FmPathList* files)
+{
+    if(!fm_config->confirm_del || fm_yes_no(parent, NULL, _("Do you want to move the selected files to trash can?"), TRUE))
     {
-        FmJob* job = fm_file_ops_job_new(FM_FILE_OP_TRASH, files);
-        fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
+        FmFileOpsJob* job = fm_file_ops_job_new(FM_FILE_OP_TRASH, files);
+        fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
     }
 }
 
-void fm_untrash_files(FmPathList* files)
+void fm_untrash_files(GtkWindow* parent, FmPathList* files)
 {
-    FmJob* job = fm_file_ops_job_new(FM_FILE_OP_UNTRASH, files);
-    fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
+    FmFileOpsJob* job = fm_file_ops_job_new(FM_FILE_OP_UNTRASH, files);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
 }
 
-static void fm_delete_files_internal(FmPathList* files)
+static void fm_delete_files_internal(GtkWindow* parent, FmPathList* files)
 {
-    FmJob* job = fm_file_ops_job_new(FM_FILE_OP_DELETE, files);
-    fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
+    FmFileOpsJob* job = fm_file_ops_job_new(FM_FILE_OP_DELETE, files);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
 }
 
-void fm_delete_files(FmPathList* files)
+void fm_delete_files(GtkWindow* parent, FmPathList* files)
 {
-    if(!fm_config->confirm_del || fm_yes_no(NULL, _("Do you want to delete the selected files?"), TRUE))
-        fm_delete_files_internal(files);
+    if(!fm_config->confirm_del || fm_yes_no(parent, NULL, _("Do you want to delete the selected files?"), TRUE))
+        fm_delete_files_internal(parent, files);
 }
 
-void fm_trash_or_delete_files(FmPathList* files)
+void fm_trash_or_delete_files(GtkWindow* parent, FmPathList* files)
 {
-    if( !fm_list_is_empty(files) )
+    if( !fm_path_list_is_empty(files) )
     {
         gboolean all_in_trash = TRUE;
         if(fm_config->use_trash)
         {
-            GList* l = fm_list_peek_head_link(files);
+            GList* l = fm_path_list_peek_head_link(files);
             for(;l;l=l->next)
             {
                 FmPath* path = FM_PATH(l->data);
@@ -522,36 +683,40 @@ void fm_trash_or_delete_files(FmPathList* files)
 
         /* files already in trash:/// should only be deleted and cannot be trashed again. */
         if(fm_config->use_trash && !all_in_trash)
-            fm_trash_files(files);
+            fm_trash_files(parent, files);
         else
-            fm_delete_files(files);
+            fm_delete_files(parent, files);
     }
 }
 
-void fm_move_or_copy_files_to(FmPathList* files, gboolean is_move)
+void fm_move_or_copy_files_to(GtkWindow* parent, FmPathList* files, gboolean is_move)
 {
-    FmPath* dest = fm_select_folder(NULL);
+    FmPath* dest = fm_select_folder(parent, NULL);
     if(dest)
     {
         if(is_move)
-            fm_move_files(files, dest);
+            fm_move_files(parent, files, dest);
         else
-            fm_copy_files(files, dest);
+            fm_copy_files(parent, files, dest);
         fm_path_unref(dest);
     }
 }
 
 
-void fm_rename_file(FmPath* file)
+void fm_rename_file(GtkWindow* parent, FmPath* file)
 {
-    GFile* gf = fm_path_to_gfile(file), *parent, *dest;
+    GFile *gf, *parent_gf, *dest;
     GError* err = NULL;
-    gchar* new_name = fm_get_user_input_rename( NULL, _("Rename File"), _("Please enter a new name:"), file->name);
+    gchar* new_name;
+    new_name = fm_get_user_input_rename(parent, _("Rename File"),
+                                        _("Please enter a new name:"),
+                                        fm_path_get_basename(file));
     if( !new_name )
         return;
-    parent = g_file_get_parent(gf);
-    dest = g_file_get_child(parent, new_name);
-    g_object_unref(parent);
+    gf = fm_path_to_gfile(file);
+    parent_gf = g_file_get_parent(gf);
+    dest = g_file_get_child(G_FILE(parent_gf), new_name);
+    g_object_unref(parent_gf);
     if(!g_file_move(gf, dest,
                 G_FILE_COPY_ALL_METADATA|
                 G_FILE_COPY_NO_FALLBACK_FOR_MOVE|
@@ -559,106 +724,44 @@ void fm_rename_file(FmPath* file)
                 NULL, /* make this cancellable later. */
                 NULL, NULL, &err))
     {
-        fm_show_error(NULL, err->message);
+        fm_show_error(parent, NULL, err->message);
         g_error_free(err);
     }
     g_object_unref(dest);
     g_object_unref(gf);
 }
 
-void fm_empty_trash()
+void fm_empty_trash(GtkWindow* parent)
 {
-    if(fm_yes_no(NULL, _("Are you sure you want to empty the trash can?"), TRUE))
+    if(fm_yes_no(parent, NULL, _("Are you sure you want to empty the trash can?"), TRUE))
     {
         FmPathList* paths = fm_path_list_new();
-        fm_list_push_tail(paths, fm_path_get_trash());
-        fm_delete_files_internal(paths);
-        fm_list_unref(paths);
+        fm_path_list_push_tail(paths, fm_path_get_trash());
+        fm_delete_files_internal(parent, paths);
+        fm_path_list_unref(paths);
     }
 }
 
-static GAppInfo* choose_app(GList* file_infos, FmMimeType* mime_type, gpointer user_data, GError** err)
+void fm_set_busy_cursor(GtkWidget* widget)
 {
-    gpointer* data = (gpointer*)user_data;
-    GtkWindow* parent = (GtkWindow*)data[0];
-    return fm_choose_app_for_mime_type(parent, mime_type, mime_type != NULL);
-}
-
-static gboolean on_launch_error(GAppLaunchContext* ctx, GError* err, gpointer user_data)
-{
-    gpointer* data = (gpointer*)user_data;
-    GtkWindow* parent = (GtkWindow*)data[0];
-    fm_show_error(parent, err->message);
-    return TRUE;
-}
-
-static gboolean on_open_folder(GAppLaunchContext* ctx, GList* folder_infos, gpointer user_data, GError** err)
-{
-    gpointer* data = (gpointer*)user_data;
-    FmLaunchFolderFunc func = (FmLaunchFolderFunc)data[1];
-    return func(ctx, folder_infos, data[2], err);
-}
-
-gboolean fm_launch_files_simple(GtkWindow* parent, GAppLaunchContext* ctx, GList* file_infos, FmLaunchFolderFunc func, gpointer user_data)
-{
-    FmFileLauncher launcher = {
-        choose_app,
-        on_open_folder,
-        on_launch_error
-    };
-    gpointer data[] = {parent, func, user_data};
-    GAppLaunchContext* _ctx = NULL;
-    gboolean ret;
-    if(ctx == NULL)
+    if(gtk_widget_get_realized(widget))
     {
-        _ctx = ctx = gdk_app_launch_context_new();
-        gdk_app_launch_context_set_screen(GDK_APP_LAUNCH_CONTEXT(ctx), parent ? gtk_widget_get_screen(GTK_WIDGET(parent)) : gdk_screen_get_default());
-        gdk_app_launch_context_set_timestamp(GDK_APP_LAUNCH_CONTEXT(ctx), gtk_get_current_event_time());
-        /* FIXME: how to handle gdk_app_launch_context_set_icon? */
+        GdkWindow* window = gtk_widget_get_window(widget);
+        GdkCursor* cursor = gdk_cursor_new(GDK_WATCH);
+        gdk_window_set_cursor(window, cursor);
     }
-    ret = fm_launch_files(ctx, file_infos, &launcher, data);
-    if(_ctx)
-        g_object_unref(_ctx);
-    return ret;
-}
-
-gboolean fm_launch_paths_simple(GtkWindow* parent, GAppLaunchContext* ctx, GList* paths, FmLaunchFolderFunc func, gpointer user_data)
-{
-    FmFileLauncher launcher = {
-        choose_app,
-        on_open_folder,
-        on_launch_error
-    };
-    gpointer data[] = {parent, func, user_data};
-    GAppLaunchContext* _ctx = NULL;
-    gboolean ret;
-    if(ctx == NULL)
+    else
     {
-        _ctx = ctx = gdk_app_launch_context_new();
-        gdk_app_launch_context_set_screen(GDK_APP_LAUNCH_CONTEXT(ctx), parent ? gtk_widget_get_screen(GTK_WIDGET(parent)) : gdk_screen_get_default());
-        gdk_app_launch_context_set_timestamp(GDK_APP_LAUNCH_CONTEXT(ctx), gtk_get_current_event_time());
-        /* FIXME: how to handle gdk_app_launch_context_set_icon? */
+        /* FIXME: how to handle this case? */
+        /* g_debug("not realized"); */
     }
-    ret = fm_launch_paths(ctx, paths, &launcher, data);
-    if(_ctx)
-        g_object_unref(_ctx);
-    return ret;
 }
 
-gboolean fm_launch_file_simple(GtkWindow* parent, GAppLaunchContext* ctx, FmFileInfo* file_info, FmLaunchFolderFunc func, gpointer user_data)
+void fm_unset_busy_cursor(GtkWidget* widget)
 {
-    gboolean ret;
-    GList* files = g_list_prepend(NULL, file_info);
-    ret = fm_launch_files_simple(parent, ctx, files, func, user_data);
-    g_list_free(files);
-    return ret;
-}
-
-gboolean fm_launch_path_simple(GtkWindow* parent, GAppLaunchContext* ctx, FmPath* path, FmLaunchFolderFunc func, gpointer user_data)
-{
-    gboolean ret;
-    GList* files = g_list_prepend(NULL, path);
-    ret = fm_launch_paths_simple(parent, ctx, files, func, user_data);
-    g_list_free(files);
-    return ret;
+    if(gtk_widget_get_realized(widget))
+    {
+        GdkWindow* window = gtk_widget_get_window(widget);
+        gdk_window_set_cursor(window, NULL);
+    }
 }
