@@ -38,20 +38,21 @@ enum
 
 struct _FmProgressDisplay
 {
-    GtkWidget* dlg;
+    GtkWindow* parent;
+    GtkDialog* dlg;
     FmFileOpsJob* job;
 
     /* private */
-    GtkWidget* icon;
-    GtkWidget* msg;
-    GtkWidget* act;
-    GtkWidget* src;
+    GtkImage* icon;
+    GtkLabel* msg;
+    GtkLabel* act;
+    GtkLabel* src;
     GtkWidget* dest;
-    GtkWidget* current;
-    GtkWidget* progress;
-    GtkWidget* remaining_time;
+    GtkLabel* current;
+    GtkProgressBar* progress;
+    GtkLabel* remaining_time;
     GtkWidget* error_pane;
-    GtkWidget* error_msg;
+    GtkTextView* error_msg;
     GtkTextBuffer* error_buf;
     GtkTextTag* bold_tag;
 
@@ -75,11 +76,10 @@ static void on_percent(FmFileOpsJob* job, guint percent, FmProgressDisplay* data
 {
     if(data->dlg)
     {
-        GTimeVal cur_time;
         char percent_text[64];
         g_snprintf(percent_text, 64, "%d %%", percent);
-        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(data->progress), (gdouble)percent/100);
-        gtk_progress_bar_set_text(GTK_PROGRESS_BAR(data->progress), percent_text);
+        gtk_progress_bar_set_fraction(data->progress, (gdouble)percent/100);
+        gtk_progress_bar_set_text(data->progress, percent_text);
 
         gdouble elapsed = g_timer_elapsed(data->timer, NULL);
         if(elapsed >= 0.5)
@@ -151,7 +151,7 @@ static FmJobErrorAction on_error(FmFileOpsJob* job, GError* err, FmJobErrorSever
     gtk_text_buffer_insert(data->error_buf, &it, err->message, -1);
     gtk_text_buffer_insert(data->error_buf, &it, "\n", 1);
 
-    if(!GTK_WIDGET_VISIBLE(data->error_pane))
+    if(!gtk_widget_get_visible(data->error_pane))
         gtk_widget_show(data->error_pane);
 
     if(data->timer)
@@ -159,10 +159,10 @@ static FmJobErrorAction on_error(FmFileOpsJob* job, GError* err, FmJobErrorSever
     return FM_JOB_CONTINUE;
 }
 
-static gint on_ask(FmFileOpsJob* job, const char* question, const char** options, FmProgressDisplay* data)
+static gint on_ask(FmFileOpsJob* job, const char* question, char* const* options, FmProgressDisplay* data)
 {
     ensure_dlg(data);
-    return fm_askv(GTK_WINDOW(data->dlg), question, options);
+    return fm_askv(GTK_WINDOW(data->dlg), NULL, question, options);
 }
 
 static void on_filename_changed(GtkEditable* entry, GtkWidget* rename)
@@ -173,7 +173,7 @@ static void on_filename_changed(GtkEditable* entry, GtkWidget* rename)
     gtk_widget_set_sensitive(rename, can_rename);
     if(can_rename)
     {
-        GtkDialog* dlg = GTK_DIALOG(gtk_widget_get_toplevel(entry));
+        GtkDialog* dlg = GTK_DIALOG(gtk_widget_get_toplevel(GTK_WIDGET(entry)));
         gtk_dialog_set_default_response(dlg, gtk_dialog_get_response_for_widget(dlg, rename));
     }
 }
@@ -181,15 +181,25 @@ static void on_filename_changed(GtkEditable* entry, GtkWidget* rename)
 static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest, char** new_name, FmProgressDisplay* data)
 {
     int res;
-    GtkBuilder* builder = gtk_builder_new();
-    GtkWidget *dlg, *src_icon, *dest_icon, *src_fi, *dest_fi, *filename, *apply_all;
+    GtkBuilder* builder;
+    GtkDialog *dlg;
+    GtkImage *src_icon, *dest_icon;
+    GtkLabel *src_fi, *dest_fi;
+    GtkEntry *filename;
+    GtkToggleButton *apply_all;
     char* tmp;
     const char* disp_size;
+    FmPath* path;
+    FmIcon* icon;
 
     /* return default operation if the user has set it */
     if(data->default_opt)
         return data->default_opt;
 
+    builder = gtk_builder_new();
+    path = fm_file_info_get_path(dest);
+    icon = fm_file_info_get_icon(src);
+
     if(data->timer)
         g_timer_stop(data->timer);
 
@@ -197,16 +207,16 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
     ensure_dlg(data);
 
     gtk_builder_add_from_file(builder, PACKAGE_UI_DIR "/ask-rename.ui", NULL);
-    dlg = (GtkWidget*)gtk_builder_get_object(builder, "dlg");
-    src_icon = (GtkWidget*)gtk_builder_get_object(builder, "src_icon");
-    src_fi = (GtkWidget*)gtk_builder_get_object(builder, "src_fi");
-    dest_icon = (GtkWidget*)gtk_builder_get_object(builder, "dest_icon");
-    dest_fi = (GtkWidget*)gtk_builder_get_object(builder, "dest_fi");
-    filename = (GtkWidget*)gtk_builder_get_object(builder, "filename");
-    apply_all = (GtkWidget*)gtk_builder_get_object(builder, "apply_all");
-    gtk_window_set_transient_for(GTK_WINDOW(dlg), data->dlg);
+    dlg = GTK_DIALOG(gtk_builder_get_object(builder, "dlg"));
+    src_icon = GTK_IMAGE(gtk_builder_get_object(builder, "src_icon"));
+    src_fi = GTK_LABEL(gtk_builder_get_object(builder, "src_fi"));
+    dest_icon = GTK_IMAGE(gtk_builder_get_object(builder, "dest_icon"));
+    dest_fi = GTK_LABEL(gtk_builder_get_object(builder, "dest_fi"));
+    filename = GTK_ENTRY(gtk_builder_get_object(builder, "filename"));
+    apply_all = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "apply_all"));
+    gtk_window_set_transient_for(GTK_WINDOW(dlg), GTK_WINDOW(data->dlg));
 
-    gtk_image_set_from_gicon(GTK_IMAGE(src_icon), src->icon->gicon, GTK_ICON_SIZE_DIALOG);
+    gtk_image_set_from_gicon(src_icon, icon->gicon, GTK_ICON_SIZE_DIALOG);
     disp_size = fm_file_info_get_disp_size(src);
     if(disp_size)
     {
@@ -222,10 +232,10 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
                               fm_file_info_get_disp_mtime(src));
     }
 
-    gtk_label_set_text(GTK_LABEL(src_fi), tmp);
+    gtk_label_set_text(src_fi, tmp);
     g_free(tmp);
 
-    gtk_image_set_from_gicon(GTK_IMAGE(dest_icon), src->icon->gicon, GTK_ICON_SIZE_DIALOG);
+    gtk_image_set_from_gicon(dest_icon, icon->gicon, GTK_ICON_SIZE_DIALOG);
     disp_size = fm_file_info_get_disp_size(dest);
     if(disp_size)
     {
@@ -241,22 +251,23 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
                               fm_file_info_get_disp_mtime(dest));
     }
 
-    gtk_label_set_text(GTK_LABEL(dest_fi), tmp);
+    gtk_label_set_text(dest_fi, tmp);
     g_free(tmp);
 
-    tmp = g_filename_display_name(dest->path->name);
-    gtk_entry_set_text(GTK_ENTRY(filename), tmp);
+    tmp = g_filename_display_name(path->name);
+    gtk_entry_set_text(filename, tmp);
     g_free(tmp);
-    g_object_set_data(G_OBJECT(filename), "old_name", dest->disp_name);
-    g_signal_connect(filename, "changed", on_filename_changed, gtk_builder_get_object(builder, "rename"));
+    tmp = (char*)fm_file_info_get_disp_name(dest); /* FIXME: cast const to char */
+    g_object_set_data(G_OBJECT(filename), "old_name", tmp);
+    g_signal_connect(filename, "changed", G_CALLBACK(on_filename_changed), gtk_builder_get_object(builder, "rename"));
 
     g_object_unref(builder);
 
-    res = gtk_dialog_run(GTK_DIALOG(dlg));
+    res = gtk_dialog_run(dlg);
     switch(res)
     {
     case RESPONSE_RENAME:
-        *new_name = g_strdup(gtk_entry_get_text(GTK_ENTRY(filename)));
+        *new_name = g_strdup(gtk_entry_get_text(filename));
         res = FM_FILE_OP_RENAME;
         break;
     case RESPONSE_OVERWRITE:
@@ -269,13 +280,13 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
         res = FM_FILE_OP_CANCEL;
     }
 
-    if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(apply_all)))
+    if(gtk_toggle_button_get_active(apply_all))
     {
         if(res == RESPONSE_OVERWRITE || res == FM_FILE_OP_SKIP)
             data->default_opt = res;
     }
 
-    gtk_widget_destroy(dlg);
+    gtk_widget_destroy(GTK_WIDGET(dlg));
 
     if(data->timer)
         g_timer_continue(data->timer);
@@ -285,34 +296,41 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
 
 static void on_finished(FmFileOpsJob* job, FmProgressDisplay* data)
 {
+    GtkWindow* parent = NULL;
     if(data->update_timeout)
     {
         g_source_remove(data->update_timeout);
         data->update_timeout = 0;
     }
 
+    /* preserve pointers that fm_progress_display_destroy() will unreference
+       as they may be requested by trash support below */
+    if(data->parent)
+        parent = g_object_ref(data->parent);
+    g_object_ref(job);
+
     if(data->dlg)
     {
         /* errors happened */
-        if(data->has_error && data->dlg)
+        if(data->has_error)
         {
-            gtk_label_set_text(GTK_LABEL(data->current), "");
-            gtk_label_set_text(GTK_LABEL(data->remaining_time), "00:00:00");
-            gtk_dialog_set_response_sensitive(GTK_DIALOG(data->dlg), GTK_RESPONSE_CANCEL, FALSE);
-            gtk_dialog_add_button(GTK_DIALOG(data->dlg), GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);
+            gtk_label_set_text(data->current, "");
+            gtk_label_set_text(data->remaining_time, "00:00:00");
+            gtk_dialog_set_response_sensitive(data->dlg, GTK_RESPONSE_CANCEL, FALSE);
+            gtk_dialog_add_button(data->dlg, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);
 
             gtk_image_set_from_stock(data->icon, GTK_STOCK_DIALOG_WARNING, GTK_ICON_SIZE_DIALOG);
 
-            gtk_widget_show(data->msg);
-            if(fm_job_is_cancelled(data->job))
+            gtk_widget_show(GTK_WIDGET(data->msg));
+            if(fm_job_is_cancelled(FM_JOB(job)))
             {
-                gtk_label_set_text(GTK_LABEL(data->msg), _("The file operation is cancelled and there are some errors."));
+                gtk_label_set_text(data->msg, _("The file operation is cancelled and there are some errors."));
                 gtk_window_set_title(GTK_WINDOW(data->dlg),
                                      _("Cancelled"));
             }
             else
             {
-                gtk_label_set_text(GTK_LABEL(data->msg), _("The file operation is finished, but there are some errors."));
+                gtk_label_set_text(data->msg, _("The file operation is finished, but there are some errors."));
                 gtk_window_set_title(GTK_WINDOW(data->dlg),
                                      _("Finished"));
             }
@@ -323,24 +341,33 @@ static void on_finished(FmFileOpsJob* job, FmProgressDisplay* data)
     }
     else
         fm_progress_display_destroy(data);
+    /* if it's not destroyed yet then it will be destroyed with dialog window */
 
     /* sepcial handling for trash
      * FIXME: need to refactor this to use a more elegant way later. */
-    if(job->type == FM_FILE_OP_TRASH)
+    if(job->type == FM_FILE_OP_TRASH) /* FIXME: direct access to job struct! */
     {
-        FmPathList* unsupported = (FmPathList*)g_object_get_data(job, "trash-unsupported");
+        FmPathList* unsupported = (FmPathList*)g_object_get_data(G_OBJECT(job), "trash-unsupported");
         /* some files cannot be trashed because underlying filesystems don't support it. */
+        g_object_unref(job);
         if(unsupported) /* delete them instead */
         {
-            if(fm_yes_no(NULL, _("Some files cannot be moved to trash can because "
+            /* FIXME: parent window might be already destroyed! */
+            if(fm_yes_no(parent, NULL,
+                        _("Some files cannot be moved to trash can because "
                         "the underlying file systems don't support this operation.\n"
                         "Do you want to delete them instead?"), TRUE))
             {
-                FmJob* job = fm_file_ops_job_new(FM_FILE_OP_DELETE, unsupported);
-                fm_file_ops_job_run_with_progress(job);
+                job = fm_file_ops_job_new(FM_FILE_OP_DELETE, unsupported);
+                fm_file_ops_job_run_with_progress(parent, job);
+                                                        /* it eats reference! */
             }
         }
     }
+    else
+        g_object_unref(job);
+    if(parent)
+        g_object_unref(parent);
 }
 
 static void on_cancelled(FmFileOpsJob* job, FmProgressDisplay* data)
@@ -364,20 +391,33 @@ static void on_response(GtkDialog* dlg, gint id, FmProgressDisplay* data)
         fm_progress_display_destroy(data);
 }
 
-static gboolean on_update_dlg(FmProgressDisplay* data)
+static gboolean on_update_dlg(gpointer user_data)
 {
+    FmProgressDisplay* data = (FmProgressDisplay*)user_data;
     if(data->old_cur_file != data->cur_file)
     {
-        gtk_label_set_text(GTK_LABEL(data->current), data->cur_file);
+        gtk_label_set_text(data->current, data->cur_file);
         data->old_cur_file = data->cur_file;
     }
     return TRUE;
 }
 
-static gboolean on_show_dlg(FmProgressDisplay* data)
+static void on_progress_dialog_destroy(gpointer user_data, GObject* dlg)
 {
+    FmProgressDisplay* data = (FmProgressDisplay*)user_data;
+
+    data->dlg = NULL; /* it's destroying right now, don't destroy it again */
+    g_object_unref(data->error_buf); /* these will be not unref if no dlg */
+    g_object_unref(data->bold_tag);
+    fm_progress_display_destroy(data);
+}
+
+static gboolean on_show_dlg(gpointer user_data)
+{
+    FmProgressDisplay* data = (FmProgressDisplay*)user_data;
     GtkBuilder* builder = gtk_builder_new();
-    GtkWidget* to, *to_label;
+    GtkLabel* to;
+    GtkWidget *to_label;
     FmPath* dest;
     const char* title = NULL;
     GtkTextTagTable* tag_table = gtk_text_tag_table_new();
@@ -385,36 +425,39 @@ static gboolean on_show_dlg(FmProgressDisplay* data)
     gtk_builder_set_translation_domain(builder, GETTEXT_PACKAGE);
     gtk_builder_add_from_file(builder, PACKAGE_UI_DIR "/progress.ui", NULL);
 
-    data->dlg = (GtkWidget*)gtk_builder_get_object(builder, "dlg");
+    data->dlg = GTK_DIALOG(gtk_builder_get_object(builder, "dlg"));
+    g_object_weak_ref(G_OBJECT(data->dlg), on_progress_dialog_destroy, data);
 
-    g_signal_connect(data->dlg, "response", on_response, data);
+    g_signal_connect(data->dlg, "response", G_CALLBACK(on_response), data);
+    /* FIXME: connect to "close" signal */
 
     to_label = (GtkWidget*)gtk_builder_get_object(builder, "to_label");
-    to = (GtkWidget*)gtk_builder_get_object(builder, "dest");
-    data->icon = (GtkWidget*)gtk_builder_get_object(builder, "icon");
-    data->msg = (GtkWidget*)gtk_builder_get_object(builder, "msg");
-    data->act = (GtkWidget*)gtk_builder_get_object(builder, "action");
-    data->src = (GtkWidget*)gtk_builder_get_object(builder, "src");
+    to = GTK_LABEL(gtk_builder_get_object(builder, "dest"));
+    data->icon = GTK_IMAGE(gtk_builder_get_object(builder, "icon"));
+    data->msg = GTK_LABEL(gtk_builder_get_object(builder, "msg"));
+    data->act = GTK_LABEL(gtk_builder_get_object(builder, "action"));
+    data->src = GTK_LABEL(gtk_builder_get_object(builder, "src"));
     data->dest = (GtkWidget*)gtk_builder_get_object(builder, "dest");
-    data->current = (GtkWidget*)gtk_builder_get_object(builder, "current");
-    data->progress = (GtkWidget*)gtk_builder_get_object(builder, "progress");
+    data->current = GTK_LABEL(gtk_builder_get_object(builder, "current"));
+    data->progress = GTK_PROGRESS_BAR(gtk_builder_get_object(builder, "progress"));
     data->error_pane = (GtkWidget*)gtk_builder_get_object(builder, "error_pane");
-    data->error_msg = (GtkWidget*)gtk_builder_get_object(builder, "error_msg");
-    data->remaining_time = (GtkWidget*)gtk_builder_get_object(builder, "remaining_time");
+    data->error_msg = GTK_TEXT_VIEW(gtk_builder_get_object(builder, "error_msg"));
+    data->remaining_time = GTK_LABEL(gtk_builder_get_object(builder, "remaining_time"));
 
     data->bold_tag = gtk_text_tag_new("bold");
     g_object_set(data->bold_tag, "weight", PANGO_WEIGHT_BOLD, NULL);
     gtk_text_tag_table_add(tag_table, data->bold_tag);
     data->error_buf = gtk_text_buffer_new(tag_table);
     g_object_unref(tag_table);
-    gtk_text_view_set_buffer(GTK_TEXT_VIEW(data->error_msg), data->error_buf);
+    gtk_text_view_set_buffer(data->error_msg, data->error_buf);
 
     g_object_unref(builder);
 
     /* set the src label */
+    /* FIXME: direct access to job struct! */
     if(data->job->srcs)
     {
-        GList* l = fm_list_peek_head_link(data->job->srcs);
+        GList* l = fm_path_list_peek_head_link(data->job->srcs);
         int i;
         char* disp;
         FmPath* path;
@@ -438,37 +481,42 @@ static gboolean on_show_dlg(FmProgressDisplay* data)
     }
 
     /* FIXME: use accessor functions instead */
+    /* FIXME: direct access to job struct! */
     switch(data->job->type)
     {
-	case FM_FILE_OP_MOVE:
+    case FM_FILE_OP_MOVE:
         title = _("Moving files");
         break;
-	case FM_FILE_OP_COPY:
+    case FM_FILE_OP_COPY:
         title = _("Copying files");
         break;
-	case FM_FILE_OP_TRASH:
+    case FM_FILE_OP_TRASH:
         title = _("Trashing files");
         break;
-	case FM_FILE_OP_DELETE:
+    case FM_FILE_OP_DELETE:
         title = _("Deleting files");
         break;
     case FM_FILE_OP_LINK:
         title = _("Creating symlinks");
         break;
-	case FM_FILE_OP_CHANGE_ATTR:
+    case FM_FILE_OP_CHANGE_ATTR:
         title = _("Changing file attributes");
         break;
+    case FM_FILE_OP_UNTRASH:
+        break;
+    case FM_FILE_OP_NONE: ;
     }
     if(title)
     {
         gtk_window_set_title(GTK_WINDOW(data->dlg), title);
-        gtk_label_set_text(GTK_LABEL(data->act), title);
+        gtk_label_set_text(data->act, title);
     }
 
-    if(dest = fm_file_ops_job_get_dest(data->job))
+    dest = fm_file_ops_job_get_dest(data->job);
+    if(dest)
     {
         char* dest_str = fm_path_display_name(dest, TRUE);
-        gtk_label_set_text(GTK_LABEL(to), dest_str);
+        gtk_label_set_text(to, dest_str);
         g_free(dest_str);
     }
     else
@@ -478,13 +526,13 @@ static gboolean on_show_dlg(FmProgressDisplay* data)
     }
 
     gtk_window_present(GTK_WINDOW(data->dlg));
-    data->update_timeout = g_timeout_add(500, (GSourceFunc)on_update_dlg, data);
+    data->update_timeout = g_timeout_add(500, on_update_dlg, data);
 
     data->delay_timeout = 0;
     return FALSE;
 }
 
-void ensure_dlg(FmProgressDisplay* data)
+static void ensure_dlg(FmProgressDisplay* data)
 {
     if(data->delay_timeout)
     {
@@ -500,15 +548,30 @@ static void on_prepared(FmFileOpsJob* job, FmProgressDisplay* data)
     data->timer = g_timer_new();
 }
 
-/* Run the file operation job with a progress dialog.
+/**
+ * fm_file_ops_job_run_with_progress
+ * @parent: parent window to show dialog over it
+ * @job: job descriptor to run
+ *
+ * Runs the file operation job with a progress dialog.
  * The returned data structure will be freed in idle handler automatically
  * when it's not needed anymore.
+ *
+ * NOTE: INCONSISTENCY: it takes a reference from job
+ *
+ * Return value: (transfer none): progress data; not usable; caller should not free it either.
  */
-FmProgressDisplay* fm_file_ops_job_run_with_progress(FmFileOpsJob* job)
+FmProgressDisplay* fm_file_ops_job_run_with_progress(GtkWindow* parent, FmFileOpsJob* job)
 {
-    FmProgressDisplay* data = g_slice_new0(FmProgressDisplay);
-    data->job = (FmFileOpsJob*)g_object_ref(job);
-    data->delay_timeout = g_timeout_add(SHOW_DLG_DELAY, (GSourceFunc)on_show_dlg, data);
+    FmProgressDisplay* data;
+
+    g_return_val_if_fail(job != NULL, NULL);
+
+    data = g_slice_new0(FmProgressDisplay);
+    data->job = job;
+    if(parent)
+        data->parent = g_object_ref(parent);
+    data->delay_timeout = g_timeout_add(SHOW_DLG_DELAY, on_show_dlg, data);
 
     g_signal_connect(job, "ask", G_CALLBACK(on_ask), data);
     g_signal_connect(job, "ask-rename", G_CALLBACK(on_ask_rename), data);
@@ -519,29 +582,32 @@ FmProgressDisplay* fm_file_ops_job_run_with_progress(FmFileOpsJob* job)
     g_signal_connect(job, "finished", G_CALLBACK(on_finished), data);
     g_signal_connect(job, "cancelled", G_CALLBACK(on_cancelled), data);
 
-	fm_job_run_async(job);
+    fm_job_run_async(FM_JOB(job));
 
     return data;
 }
 
-void fm_progress_display_destroy(FmProgressDisplay* data)
+static void fm_progress_display_destroy(FmProgressDisplay* data)
 {
-    if(data->job)
-    {
-        fm_job_cancel(FM_JOB(data->job));
+    g_signal_handlers_disconnect_by_func(data->job, on_cancelled, data);
 
-        g_signal_handlers_disconnect_by_func(data->job, on_ask, data);
-        g_signal_handlers_disconnect_by_func(data->job, on_ask_rename, data);
-        g_signal_handlers_disconnect_by_func(data->job, on_error, data);
-        g_signal_handlers_disconnect_by_func(data->job, on_cur_file, data);
-        g_signal_handlers_disconnect_by_func(data->job, on_percent, data);
-        g_signal_handlers_disconnect_by_func(data->job, on_finished, data);
+    fm_job_cancel(FM_JOB(data->job));
 
-        g_object_unref(data->job);
+    g_signal_handlers_disconnect_by_func(data->job, on_ask, data);
+    g_signal_handlers_disconnect_by_func(data->job, on_ask_rename, data);
+    g_signal_handlers_disconnect_by_func(data->job, on_error, data);
+    g_signal_handlers_disconnect_by_func(data->job, on_prepared, data);
+    g_signal_handlers_disconnect_by_func(data->job, on_cur_file, data);
+    g_signal_handlers_disconnect_by_func(data->job, on_percent, data);
+    g_signal_handlers_disconnect_by_func(data->job, on_finished, data);
 
-        if(data->timer)
-            g_timer_destroy(data->timer);
-    }
+    g_object_unref(data->job);
+
+    if(data->timer)
+        g_timer_destroy(data->timer);
+
+    if(data->parent)
+        g_object_unref(data->parent);
 
     g_free(data->cur_file);
 
@@ -553,9 +619,10 @@ void fm_progress_display_destroy(FmProgressDisplay* data)
 
     if(data->dlg)
     {
+        g_object_weak_unref(G_OBJECT(data->dlg), on_progress_dialog_destroy, data);
         g_object_unref(data->error_buf);
         g_object_unref(data->bold_tag);
-        gtk_widget_destroy(data->dlg);
+        gtk_widget_destroy(GTK_WIDGET(data->dlg));
     }
 
     g_slice_free(FmProgressDisplay, data);
