@@ -42,11 +42,12 @@ static const char query[]=
     G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE","
     G_FILE_ATTRIBUTE_ID_FILESYSTEM;
 
-static void progress_cb(goffset cur, goffset total, FmFileOpsJob* job);
+static void progress_cb(goffset cur, goffset total, gpointer job);
 
-gboolean _fm_file_ops_job_check_paths(FmFileOpsJob* job, GFile* src, GFileInfo* src_inf, GFile* dest)
+static gboolean _fm_file_ops_job_check_paths(FmFileOpsJob* job, GFile* src, GFileInfo* src_inf, GFile* dest)
 {
     GError* err = NULL;
+    FmJob* fmjob = FM_JOB(job);
     if(job->type == FM_FILE_OP_MOVE && g_file_equal(src, dest))
     {
         err = g_error_new_literal(G_IO_ERROR, G_IO_ERROR_FAILED,
@@ -66,29 +67,29 @@ gboolean _fm_file_ops_job_check_paths(FmFileOpsJob* job, GFile* src, GFileInfo*
     }
     if(err)
     {
-        if(!fm_job_is_cancelled(FM_JOB(job)))
+        if(!fm_job_is_cancelled(fmjob))
         {
             fm_file_ops_job_emit_cur_file(job, g_file_info_get_display_name(src_inf));
-            fm_job_emit_error(FM_JOB(job), err, FM_JOB_ERROR_CRITICAL);
+            fm_job_emit_error(fmjob, err, FM_JOB_ERROR_CRITICAL);
         }
         g_error_free(err);
     }
     return (err == NULL);
 }
 
-gboolean _fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
+static gboolean _fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
 {
     /* FIXME: prevent copying to self or copying parent dir to child. */
     gboolean ret = FALSE;
     gboolean delete_src = FALSE;
     GError* err = NULL;
-    gboolean is_virtual;
     GFileType type;
     guint64 size;
     GFile* new_dest = NULL;
     GFileCopyFlags flags;
     FmJob* fmjob = FM_JOB(job);
     guint32 mode;
+    gboolean skip_dir_content = FALSE;
 
     if( G_LIKELY(inf) )
         g_object_ref(inf);
@@ -120,7 +121,6 @@ _retry_query_src_info:
     /* showing currently processed file. */
     fm_file_ops_job_emit_cur_file(job, g_file_info_get_display_name(inf));
 
-    is_virtual = g_file_info_get_attribute_boolean(inf, G_FILE_ATTRIBUTE_STANDARD_IS_VIRTUAL);
     type = g_file_info_get_file_type(inf);
 
     size = g_file_info_get_size(inf);
@@ -135,25 +135,23 @@ _retry_query_src_info:
         {
             GFileEnumerator* enu;
             gboolean dir_created = FALSE;
-        _retry_mkdir:
-            if( !fm_job_is_cancelled(fmjob) &&
+_retry_mkdir:
+            if( !fm_job_is_cancelled(fmjob) && !job->skip_dir_content &&
                 !g_file_make_directory(dest, fm_job_get_cancellable(fmjob), &err) )
             {
                 if(err->domain == G_IO_ERROR && err->code == G_IO_ERROR_EXISTS)
                 {
+                    GFile* dest_cp = new_dest;
                     FmFileOpOption opt = 0;
                     g_error_free(err);
                     err = NULL;
 
-                    g_object_ref(dest); /*if dest == new_dest, it will become invalid when we unref new_dest */
-                    if(new_dest)
-                    {
-                        g_object_unref(new_dest);
-                        new_dest = NULL;
-                    }
+                    new_dest = NULL;
                     opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
-                    g_object_unref(dest);
-
+                    if(!new_dest) /* restoring status quo */
+                        new_dest = dest_cp;
+                    else if(dest_cp) /* we got new new_dest, forget old one */
+                        g_object_unref(dest_cp);
                     switch(opt)
                     {
                     case FM_FILE_OP_RENAME:
@@ -164,18 +162,16 @@ _retry_query_src_info:
                         /* when a dir is skipped, we need to know its total size to calculate correct progress */
                         job->finished += size;
                         fm_file_ops_job_emit_percent(job);
-                        job->skip_dir_content = TRUE;
-                        ret = FALSE;
+                        job->skip_dir_content = skip_dir_content = TRUE;
                         dir_created = TRUE; /* pretend that dir creation succeeded */
                         break;
                     case FM_FILE_OP_OVERWRITE:
-                        ret = TRUE;
                         dir_created = TRUE; /* pretend that dir creation succeeded */
                         break;
                     case FM_FILE_OP_CANCEL:
-                        fm_job_cancel(FM_JOB(job));
-                        ret = FALSE;
+                        fm_job_cancel(fmjob);
                         break;
+                    case FM_FILE_OP_SKIP_ERROR: ; /* FIXME */
                     }
                 }
                 else if(!fm_job_is_cancelled(fmjob))
@@ -185,7 +181,6 @@ _retry_query_src_info:
                     err = NULL;
                     if(act == FM_JOB_RETRY)
                         goto _retry_mkdir;
-                    ret = FALSE;
                 }
                 job->finished += size;
                 fm_file_ops_job_emit_percent(job);
@@ -193,15 +188,15 @@ _retry_query_src_info:
             else
             {
                 /* chmod the newly created dir properly */
-                if(!fm_job_is_cancelled(fmjob))
+                if(!fm_job_is_cancelled(fmjob) && !job->skip_dir_content)
                 {
                     if(mode)
                     {
-                    _retry_chmod_for_dir:
+_retry_chmod_for_dir:
                         mode |= (S_IRUSR|S_IWUSR); /* ensure we have rw permission to this file. */
                         if( !g_file_set_attribute_uint32(dest, G_FILE_ATTRIBUTE_UNIX_MODE,
                                                          mode, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
-                                                         fm_job_get_cancellable(FM_JOB(job)), &err) )
+                                                         fm_job_get_cancellable(fmjob), &err) )
                         {
                             FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
                             g_error_free(err);
@@ -212,7 +207,6 @@ _retry_query_src_info:
                         }
                     }
                     dir_created = TRUE;
-                    ret = TRUE;
                 }
                 job->finished += size;
                 fm_file_ops_job_emit_percent(job);
@@ -222,19 +216,23 @@ _retry_query_src_info:
             }
 
             if(!dir_created) /* if target dir is not created, don't copy dir content */
-                job->skip_dir_content = TRUE;
+            {
+                if(!job->skip_dir_content)
+                    job->skip_dir_content = skip_dir_content = TRUE;
+            }
 
             /* the dest dir is created. let's copy its content. */
             /* FIXME: handle the case when the dir cannot be created. */
-            if(!fm_job_is_cancelled(fmjob))
+            else if(!fm_job_is_cancelled(fmjob))
             {
-            _retry_enum_children:
+_retry_enum_children:
                 enu = g_file_enumerate_children(src, query,
                                     0, fm_job_get_cancellable(fmjob), &err);
                 if(enu)
                 {
                     int n_children = 0;
                     int n_copied = 0;
+                    ret = TRUE;
                     while( !fm_job_is_cancelled(fmjob) )
                     {
                         inf = g_file_enumerator_next_file(enu, fm_job_get_cancellable(fmjob), &err);
@@ -270,6 +268,8 @@ _retry_query_src_info:
 
                                 if(ret2)
                                     ++n_copied;
+                                else
+                                    ret = FALSE;
                             }
                             g_object_unref(inf);
                         }
@@ -297,12 +297,11 @@ _retry_query_src_info:
                                         if(n_children != n_copied)
                                         {
                                             /* if the copy actions are skipped deliberately, it's ok */
-                                            if(job->skip_dir_content)
-                                                ret = TRUE;
+                                            if(!job->skip_dir_content)
+                                                ret = FALSE;
                                         }
                                     }
-                                    else
-                                        ret = FALSE;
+                                    /* else job->skip_dir_content is TRUE */
                                 }
                                 break;
                             }
@@ -320,11 +319,10 @@ _retry_query_src_info:
                         goto _retry_enum_children;
                 }
             }
-            else /* the operation is cancelled */
-            {
-                ret = FALSE;
-            }
-            job->skip_dir_content = FALSE;
+            if(job->skip_dir_content)
+                delete_src = FALSE;
+            if(skip_dir_content)
+                job->skip_dir_content = FALSE;
         }
         break;
 
@@ -355,7 +353,7 @@ _retry_query_src_info:
                         /* FIXME: error handling */
                     }
                 }
-                /* FIXME: how about blcok device, char device, and socket? */
+                /* FIXME: how about block device, char device, and socket? */
             }
             else
             {
@@ -367,7 +365,7 @@ _retry_query_src_info:
 
     default:
         flags = G_FILE_COPY_ALL_METADATA|G_FILE_COPY_NOFOLLOW_SYMLINKS;
-    _retry_copy:
+_retry_copy:
         if( !g_file_copy(src, dest, flags, fm_job_get_cancellable(fmjob),
                          progress_cb, fmjob, &err) )
         {
@@ -376,18 +374,17 @@ _retry_query_src_info:
             /* handle existing files */
             if(err->domain == G_IO_ERROR && err->code == G_IO_ERROR_EXISTS)
             {
+                GFile* dest_cp = new_dest;
                 FmFileOpOption opt = 0;
                 g_error_free(err);
                 err = NULL;
 
-                g_object_ref(dest); /*if dest == new_dest, it will become invalid when we unref new_dest */
-                if(new_dest)
-                {
-                    g_object_unref(new_dest);
-                    new_dest = NULL;
-                }
+                new_dest = NULL;
                 opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
-                g_object_unref(dest);
+                if(!new_dest) /* restoring status quo */
+                    new_dest = dest_cp;
+                else if(dest_cp) /* we got new new_dest, forget old one */
+                    g_object_unref(dest_cp);
                 switch(opt)
                 {
                 case FM_FILE_OP_RENAME:
@@ -399,17 +396,19 @@ _retry_query_src_info:
                     goto _retry_copy;
                     break;
                 case FM_FILE_OP_CANCEL:
-                    fm_job_cancel(FM_JOB(job));
-                    ret = FALSE;
+                    fm_job_cancel(fmjob);
                     break;
                 case FM_FILE_OP_SKIP:
                     ret = TRUE;
                     delete_src = FALSE; /* don't delete source file. */
                     break;
+                case FM_FILE_OP_SKIP_ERROR: ; /* FIXME */
                 }
             }
-            if(err)
+            else
             {
+                gboolean is_no_space = (err->domain == G_IO_ERROR &&
+                                        err->code == G_IO_ERROR_NO_SPACE);
                 FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
                 g_error_free(err);
                 err = NULL;
@@ -418,6 +417,9 @@ _retry_query_src_info:
                     job->current_file_finished = 0;
                     goto _retry_copy;
                 }
+                /* FIXME: ask to leave partial content? */
+                if(is_no_space)
+                    g_file_delete(dest, fm_job_get_cancellable(fmjob), NULL);
                 ret = FALSE;
                 delete_src = FALSE;
             }
@@ -441,7 +443,7 @@ _retry_query_src_info:
     /* if this is a cross-device move operation, delete source files. */
     /* ret == TRUE means the copy is successful. */
     if( !fm_job_is_cancelled(fmjob) && ret && delete_src )
-        ret = _fm_file_ops_job_delete_file(FM_JOB(job), src, inf); /* delete the source file. */
+        ret = _fm_file_ops_job_delete_file(fmjob, src, inf); /* delete the source file. */
 
     if(new_dest)
         g_object_unref(new_dest);
@@ -490,21 +492,21 @@ _retry_query_src_info:
 
         /* showing currently processed file. */
         fm_file_ops_job_emit_cur_file(job, g_file_info_get_display_name(inf));
-        _retry_move:
+_retry_move:
         if( !g_file_move(src, dest, flags, fm_job_get_cancellable(fmjob), progress_cb, job, &err))
         {
             flags &= ~G_FILE_COPY_OVERWRITE;
             if(err->domain == G_IO_ERROR && err->code == G_IO_ERROR_EXISTS)
             {
+                GFile* dest_cp = new_dest;
                 FmFileOpOption opt = 0;
-                g_object_ref(dest); /*if dest == new_dest, it will become invalid when we unref new_dest */
-                if(new_dest)
-                {
-                    g_object_unref(new_dest);
-                    new_dest = NULL;
-                }
+
+                new_dest = NULL;
                 opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
-                g_object_unref(dest);
+                if(!new_dest) /* restoring status quo */
+                    new_dest = dest_cp;
+                else if(dest_cp) /* we got new new_dest, forget old one */
+                    g_object_unref(dest_cp);
 
                 g_error_free(err);
                 err = NULL;
@@ -520,13 +522,13 @@ _retry_query_src_info:
                     if(g_file_info_get_file_type(inf) == G_FILE_TYPE_DIRECTORY) /* merge dirs */
                     {
                         GFileEnumerator* enu = g_file_enumerate_children(src, query, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
-                                                                        fm_job_get_cancellable(FM_JOB(job)), &err);
+                                                                        fm_job_get_cancellable(fmjob), &err);
                         if(enu)
                         {
                             GFileInfo* child_inf;
-                            while(!fm_job_is_cancelled(FM_JOB(job)))
+                            while(!fm_job_is_cancelled(fmjob))
                             {
-                                child_inf = g_file_enumerator_next_file(enu, fm_job_get_cancellable(FM_JOB(job)), &err);
+                                child_inf = g_file_enumerator_next_file(enu, fm_job_get_cancellable(fmjob), &err);
                                 if(child_inf)
                                 {
                                     GFile* child = g_file_get_child(src, g_file_info_get_name(child_inf));
@@ -540,7 +542,7 @@ _retry_query_src_info:
                                 {
                                     if(err) /* error */
                                     {
-                                        FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+                                        fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
                                         g_error_free(err);
                                         err = NULL;
                                     }
@@ -554,7 +556,8 @@ _retry_query_src_info:
                         }
                         else
                         {
-                            FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+                            /*FmJobErrorAction act = */
+                            fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
                             g_error_free(err);
                             err = NULL;
                             /* if(act == FM_JOB_RETRY)
@@ -562,9 +565,9 @@ _retry_query_src_info:
                         }
 
                         /* remove source dir after its content is merged with destination dir */
-                        if(!g_file_delete(src, fm_job_get_cancellable(FM_JOB(job)), &err))
+                        if(!g_file_delete(src, fm_job_get_cancellable(fmjob), &err))
                         {
-                            FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+                            fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
                             g_error_free(err);
                             err = NULL;
                             /* FIXME: should this be recoverable? */
@@ -574,12 +577,13 @@ _retry_query_src_info:
                         goto _retry_move;
                     break;
                 case FM_FILE_OP_CANCEL:
-                    fm_job_cancel(FM_JOB(job));
+                    fm_job_cancel(fmjob);
                     ret = FALSE;
                     break;
                 case FM_FILE_OP_SKIP:
                     ret = TRUE;
                     break;
+                case FM_FILE_OP_SKIP_ERROR: ; /* FIXME */
                 }
             }
             if(err)
@@ -621,8 +625,9 @@ _retry_query_src_info:
     return ret;
 }
 
-void progress_cb(goffset cur, goffset total, FmFileOpsJob* job)
+static void progress_cb(goffset cur, goffset total, gpointer data)
 {
+    FmFileOpsJob* job = FM_FILE_OPS_JOB(data);
     job->current_file_finished = cur;
     /* update progress */
     fm_file_ops_job_emit_percent(job);
@@ -632,7 +637,7 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
 {
     gboolean ret = TRUE;
     GFile *dest_dir;
-    GFileMonitor *dest_mon;
+    GFileMonitor *old_mon, *dest_mon;
     GList* l;
     FmJob* fmjob = FM_JOB(job);
     /* prepare the job, count total work needed with FmDeepCountJob */
@@ -648,7 +653,7 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
         return FALSE;
     }
     g_object_unref(dc);
-    g_debug("total size to copy: %llu", job->total);
+    g_debug("total size to copy: %llu", (long long unsigned int)job->total);
 
     dest_dir = fm_path_to_gfile(job->dest);
     /* get dummy file monitors for non-native filesystems */
@@ -656,15 +661,16 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
         dest_mon = NULL;
     else
     {
+        old_mon = job->dest_folder_mon;
         dest_mon = fm_monitor_lookup_dummy_monitor(dest_dir);
         job->dest_folder_mon = dest_mon;
     }
 
     fm_file_ops_job_emit_prepared(job);
 
-    for(l = fm_list_peek_head_link(job->srcs); !fm_job_is_cancelled(fmjob) && l; l=l->next)
+    for(l = fm_path_list_peek_head_link(job->srcs); !fm_job_is_cancelled(fmjob) && l; l=l->next)
     {
-        FmPath* path = (FmPath*)l->data;
+        FmPath* path = FM_PATH(l->data);
         GFile* src = fm_path_to_gfile(path);
         GFile* dest = g_file_get_child(dest_dir, path->name);
 
@@ -681,15 +687,15 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
     if(dest_mon)
     {
         g_object_unref(dest_mon);
-        job->dest_folder_mon = NULL;
+        job->dest_folder_mon = old_mon;
     }
-    return TRUE;
+    return ret;
 }
 
 gboolean _fm_file_ops_job_move_run(FmFileOpsJob* job)
 {
     GFile *dest_dir;
-    GFileMonitor *dest_mon;
+    GFileMonitor *old_mon, *dest_mon, *old_src_mon;
     GFileInfo* inf;
     GList* l;
     GError* err = NULL;
@@ -715,7 +721,9 @@ _retry_query_dest_info:
     }
     else
     {
-        FmJobErrorAction act = fm_job_emit_error(FM_JOB(job), err, FM_JOB_ERROR_MODERATE);
+        FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+        g_error_free(err);
+        err = NULL;
         if(act == FM_JOB_RETRY)
             goto _retry_query_dest_info;
         else
@@ -738,40 +746,38 @@ _retry_query_dest_info:
         return FALSE;
     }
     g_object_unref(dc);
-    g_debug("total size to move: %llu, dest_fs: %s", job->total, job->dest_fs_id);
+    g_debug("total size to move: %llu, dest_fs: %s",
+            (long long unsigned int)job->total, job->dest_fs_id);
 
     /* get dummy file monitors for non-native filesystems */
     if( g_file_is_native(dest_dir) )
         dest_mon = NULL;
     else
     {
+        old_mon = job->dest_folder_mon;
         dest_mon = fm_monitor_lookup_dummy_monitor(dest_dir);
         job->dest_folder_mon = dest_mon;
     }
 
     fm_file_ops_job_emit_prepared(job);
 
-    for(l = fm_list_peek_head_link(job->srcs); !fm_job_is_cancelled(fmjob) && l; l=l->next)
+    old_src_mon = job->src_folder_mon;
+    for(l = fm_path_list_peek_head_link(job->srcs); !fm_job_is_cancelled(fmjob) && l; l=l->next)
     {
-        GFileMonitor *src_mon;
-        FmPath* path = (FmPath*)l->data;
+        FmPath* path = FM_PATH(l->data);
         GFile* src = fm_path_to_gfile(path);
         GFile* dest = g_file_get_child(dest_dir, path->name);
 
         /* get dummy file monitors for non-native filesystems */
-        if( g_file_is_native(src) )
-            src_mon = NULL;
-        else
+        job->src_folder_mon = NULL;
+        if(!g_file_is_native(src))
         {
             GFile* src_dir = g_file_get_parent(src);
             if(src_dir)
             {
-                src_mon = fm_monitor_lookup_dummy_monitor(src_dir);
-                job->src_folder_mon = src_mon;
+                job->src_folder_mon = fm_monitor_lookup_dummy_monitor(src_dir);
                 g_object_unref(src_dir);
             }
-            else
-                job->src_folder_mon = src_mon = NULL;
         }
 
         if(!_fm_file_ops_job_move_file(job, src, NULL, dest))
@@ -779,21 +785,19 @@ _retry_query_dest_info:
         g_object_unref(src);
         g_object_unref(dest);
 
-        if(src_mon)
-        {
-            g_object_unref(src_mon);
-            job->src_folder_mon = NULL;
-        }
+        if(job->src_folder_mon)
+            g_object_unref(job->src_folder_mon);
 
         if(!ret)
             break;
     }
+    job->src_folder_mon = old_src_mon;
 
     g_object_unref(dest_dir);
     if(dest_mon)
     {
         g_object_unref(dest_mon);
-        job->dest_folder_mon = NULL;
+        job->dest_folder_mon = old_mon;
     }
     return ret;
 }
